unit freereverb;
{
  Unit: Reverb for either KOL or VCL Delphi and Freepascal 1.9.x,
  probably Kylix.
  purpose: Based on the Freeverb (C++) design by Jezar[at]dreampoint.co.uk
  Author: Thaddy de Koning, thaddy[at]thaddy.com
  Copyright: Original in C++ 2000, Jezar
  Delphi and Freepascal version 2003,
  Basm  added 2004, Thaddy de Koning
  Use as you like, copyrighted freeware.

  Remarks: Removed original skip for interleave.
  Comb and Allpass filter processing rewritten in BASM.
  It is as faithfull a translation as I could manage.
  Sound is exactly the same as FreeVerb 3.
  I have writen it to be as portable as my Pascals go
  Although not tested, it should work in BCB6+ and Kylix too.
  It works with Delphi 6/7 and Freepascal 1.9.X under both linux
  and windows.
  There are no dependencies on the windows unit.

  2014-05-16 : Changed to be used in WREN by Blue Hell to work on a single
  sample pair instead of a buffer. This was downloaded from :
  http://members.chello.nl/t.koning8/kolindex.htm .
  It will not work on all platforms anymore.
}

interface


uses

  Classes;


//
// Reverb model tuning values, taken from original algoritm by Jezar
//

const

  Kdenorm : single = 1.0e-23;

  numcombs     = 8;
  numallpasses = 4;
  muted        = 0;
  fixedgain    = 0.015;
  scalewet     = 3;
  scaledry     = 2;
  scaledamp    = 0.4;
  scaleroom    = 0.28;
  offsetroom   = 0.7;
  initialroom  = 0.5;
  initialdamp  = 0.5;
  initialwet   = 1 / scalewet;
  initialdry   = 1 / scaledry;
  initialwidth = 1;
  initialmode  = 0;
  freezemode   = 0.5;
  stereospread = 23;

  // These values assume 44.1KHz sample rate
  // they will probably be OK for 48KHz sample rate
  // but would need scaling for 96KHz (or other) sample rates.
  // The values were obtained by listening tests.

  combtuningL1    = 1116;
  combtuningR1    = 1116 + stereospread;
  combtuningL2    = 1188;
  combtuningR2    = 1188 + stereospread;
  combtuningL3    = 1277;
  combtuningR3    = 1277 + stereospread;
  combtuningL4    = 1356;
  combtuningR4    = 1356 + stereospread;
  combtuningL5    = 1422;
  combtuningR5    = 1422 + stereospread;
  combtuningL6    = 1491;
  combtuningR6    = 1491 + stereospread;
  combtuningL7    = 1557;
  combtuningR7    = 1557 + stereospread;
  combtuningL8    = 1617;
  combtuningR8    = 1617 + stereospread;
  allpasstuningL1 = 556;
  allpasstuningR1 = 556 + stereospread;
  allpasstuningL2 = 441;
  allpasstuningR2 = 441 + stereospread;
  allpasstuningL3 = 341;
  allpasstuningR3 = 341 + stereospread;
  allpasstuningL4 = 225;
  allpasstuningR4 = 225 + stereospread;

type

  TAllpass = class( TObject)
  private
    feedback : single;
    buffer   : pointer;
    bufsize  : integer;
    bufidx   : integer;
  public
    constructor Create( Buffersize : integer);                                                                  virtual;
    destructor  Destroy;                                                                                       override;
    function    process( const input : single) : single;                                                       register;
    procedure   mute;
    procedure   setfeedback( Value : single);
    function    getfeedback: single;
  end;

  Tcomb = class( TObject)
  private
    feedback    : single;
    filterstore : single;
    damp1       : single;
    damp2       : single;
    buffer      : pointer;
    bufsize     : integer;
    bufidx      : integer;
  public
    constructor Create( Buffersize : integer);                                                                  virtual;
    destructor  Destroy;                                                                                       override;
    function    process( const input: single): single;                                                         register;
    procedure   mute;
    procedure   setdamp    ( Value : single);
    function    getdamp    : single;
    procedure   setfeedback( Value : single);
    function    getfeedback: single;
  end;


  TReverb = class( TObject)
  private
    gain      : single;
    roomsize  : single;
    roomsize1 : single;
    damp      : single;
    damp1     : single;
    wet       : single;
    wet1      : single;
    wet2      : single;
    dry       : single;
    Width     : single;
    mode      : single;
    combL     : array[ 0 .. numcombs     - 1] of Tcomb;
    combR     : array[ 0 .. numcombs     - 1] of Tcomb;
    allpassL  : array[ 0 .. numallpasses - 1] of TAllpass;
    allpassR  : array[ 0 .. numallpasses - 1] of TAllpass;
  protected
    procedure update;
  public
    constructor Create;
    destructor  Destroy;                                                                                       override;
    procedure   mute;
    procedure   processmix( const inputL : single; const inputR : single; var outputL : single; var outputR : single);
    procedure   processreplace( const inputL : single; const inputR : single; var outputL : single; var outputR : single);
    procedure   setroomsize( Value : single);
    function    getroomsize: single;
    procedure   setdamp    ( Value : single);
    function    getdamp    : single;
    procedure   setwet     ( Value : single);
    function    getwet     : single;
    procedure   setdry     ( Value : single);
    function    getdry     : single;
    procedure   setwidth   ( Value : single);
    function    getwidth   : single;
    procedure   setmode    ( Value : single);
    function    getmode    : single;
  end;


implementation


{ ========
  TAllpass = class( TObject)
  private
    feedback : single;
    buffer   : pointer;
    bufsize  : integer;
    bufidx   : integer;
  public
 }

    constructor TAllpass.Create( Buffersize: integer);
    begin
      inherited Create;
      bufsize := Buffersize * 4;
      buffer  := AllocMem(bufsize);
      bufidx  := 0;
    end;

    destructor TAllpass.Destroy;
    begin
      freemem( buffer);
      inherited;
    end;

    function TAllpass.getfeedback: single;
    begin
      Result := feedback;
    end;

    procedure TAllpass.mute;
    begin
      Fillchar( buffer^, bufsize, 0);
    end;

    { I really don't know if this is all as fast as can be,
      but it beats Delphi's compiler generated code hands down,
      Thaddy }
    function TAllpass.process( const input: single): single;
    asm
      mov  ecx, [ eax].buffer                 // buffer start in ecx
      mov  edx, [ eax].Bufidx                 // buffer index in edx
      fld  input

      // This checks for very small values that can cause a processor
      // to switch in extra precision mode, which is expensive.
      // Since such small values are irrelevant to audio, avoid this.
      // The code is equivalent to the C inline macro by Jezar
      // This is the same spot where the original C macro appears
      test dword ptr [ ecx + edx], $7F800000  // test if denormal
      jnz @Normal
      mov dword ptr [ ecx + edx], 0           // if so, zero out
    @normal:

      fld  [ ecx + edx].Single                // load current sample from buffer
      fsub st( 0), st( 1)                     // subtract input sample
      // NOT fsub, because delphi 7 translates that into fsubp!
      fxch                                    // this is a zero cycle operant,
      // just renames the stack internally
      fmul [ eax].feedback                    // multiply stored sample with feedback
      fadd input                              // and add the input
      fstp [ ecx + edx].Single;               // store at the current sample pos
      add  edx, 4                             // increment sample position
      cmp  edx, [ eax].BufSize;               // are we at end of buffer?
      jb   @OK
      xor  edx, edx                           // if so, reset buffer index
    @OK:
      mov  [ eax].bufidx, edx                 // and store new index,
      // result already in st(0),
      // hence the fxch
    end;

    procedure TAllpass.setfeedback( Value : single);
    begin
      feedback := Value;
    end;

{ ========
  Tcomb = class( TObject)
  private
    feedback    : single;
    filterstore : single;
    damp1       : single;
    damp2       : single;
    buffer      : pointer;
    bufsize     : integer;
    bufidx      : integer;
  public
}

    constructor Tcomb.Create( Buffersize : integer);
    begin
      inherited Create;
      bufsize     := Buffersize * 4;
      buffer      := AllocMem( bufsize);
      filterstore := 0;
      bufidx      := 0;
    end;

    destructor Tcomb.Destroy;
    begin
      freemem( buffer);
      inherited;
    end;

    function Tcomb.getdamp : single;
    begin
      Result := damp1;
    end;

    function Tcomb.getfeedback : single;
    begin
      Result := feedback;
    end;

    procedure Tcomb.mute;
    begin
      Fillchar( buffer^, bufsize, 0);
    end;

    { I really don't know if this is all as fast as can be,
      but it beats Delphi's compiler generated code hands down,
      Thaddy }

    function Tcomb.process(const input : single) : single;
    asm
      mov   ecx, [ eax].Buffer                       // buffer start in ecx
      mov   edx, [ eax].Bufidx                       // buffer index in edx

      // This checks for very small values that can cause a processor
      // to switch in extra precision mode, which is expensive.
      // Since such small values are irrelevant to audio, avoid this.
      // This is the same spot where the original C macro appears
      test  dword ptr [ ecx + edx], $7F800000        // test if denormal
      jnz   @Normal
      mov   dword ptr [ ecx + edx], 0                // if so, zero out
    @normal:

      fld   [ ecx + edx].Single;                     // load sample from buffer
      fld   st( 0)                                   // duplicate on the stack
      fmul  [ eax].damp2                             // multiply with damp2
      fld   [ eax].filterstore;                      // load stored filtered sample
      fmul  [ eax].damp1                             // multiply with damp1
      faddp
      // fadd  Kdenorm
      fst   [ eax].filterstore                       // store it back

      // This checks for very small values that can cause a processor
      // to switch in extra precision mode, which is expensive.
      // Since such small values are irrelevant to audio, avoid this.
      // This is the same spot where the original C macro appears
      test  dword ptr [ eax].filterstore, $7F800000  // test if denormal
      jnz   @Normal2
      mov   dword ptr [ eax].filterstore, 0          // if so, zero out
    @normal2:

      fmul  [ eax].feedback                          // multiply with feedback
      fadd  input                                    // and add to input sample

      // fadd  Kdenorm
      fstp  [ ecx + edx].Single                      // store at current buffer pos
      add   edx, 4                                   // update buffer index
      cmp   edx, [ eax].BufSize;                     // end of buffer reached?
      jb    @OK

      xor   edx, edx                                 // if so, reset buffer index
    @OK:
      mov  [ eax].bufidx, edx                        // and store new index.
      // result already in st(0),
      // hence duplicate
    end;

    procedure Tcomb.setdamp(Value : single);
    begin
      damp1 := Value;
      damp2 := 1 - Value;
    end;

    procedure Tcomb.setfeedback(Value : single);
    begin
      feedback := Value;
    end;

    constructor TReverb.Create;
    begin
      inherited;

      combL   [ 0] := Tcomb   .Create( combtuningL1   );
      combR   [ 0] := Tcomb   .Create( combtuningR1   );
      combL   [ 1] := Tcomb   .Create( combtuningL2   );
      combR   [ 1] := Tcomb   .Create( combtuningR2   );
      combL   [ 2] := Tcomb   .Create( combtuningL3   );
      combR   [ 2] := Tcomb   .Create( combtuningR3   );
      combL   [ 3] := Tcomb   .Create( combtuningL4   );
      combR   [ 3] := Tcomb   .Create( combtuningR4   );
      combL   [ 4] := Tcomb   .Create( combtuningL5   );
      combR   [ 4] := Tcomb   .Create( combtuningR5   );
      combL   [ 5] := Tcomb   .Create( combtuningL6   );
      combR   [ 5] := Tcomb   .Create( combtuningR6   );
      combL   [ 6] := Tcomb   .Create( combtuningL7   );
      combR   [ 6] := Tcomb   .Create( combtuningR7   );
      combL   [ 7] := Tcomb   .Create( combtuningL8   );
      combR   [ 7] := Tcomb   .Create( combtuningR8   );

      allpassL[ 0] := TAllpass.Create( allpasstuningL1);
      allpassR[ 0] := TAllpass.Create( allpasstuningR1);
      allpassL[ 1] := TAllpass.Create( allpasstuningL2);
      allpassR[ 1] := TAllpass.Create( allpasstuningR2);
      allpassL[ 2] := TAllpass.Create( allpasstuningL3);
      allpassR[ 2] := TAllpass.Create( allpasstuningR3);
      allpassL[ 3] := TAllpass.Create( allpasstuningL4);
      allpassR[ 3] := TAllpass.Create( allpasstuningR4);

      // Set default values

      allpassL[ 0].setfeedback( 0.5);
      allpassR[ 0].setfeedback( 0.5);
      allpassL[ 1].setfeedback( 0.5);
      allpassR[ 1].setfeedback( 0.5);
      allpassL[ 2].setfeedback( 0.5);
      allpassR[ 2].setfeedback( 0.5);
      allpassL[ 3].setfeedback( 0.5);
      allpassR[ 3].setfeedback( 0.5);

      setwet     ( initialwet  );
      setroomsize( initialroom );
      setdry     ( initialdry  );
      setdamp    ( initialdamp );
      setwidth   ( initialwidth);
      setmode    ( initialmode );
      mute;
    end;

    destructor TReverb.Destroy;
    var
      i : integer;
    begin
      for i := 0 to 3
      do begin
        allpassL[ i].Free;
        allpassR[ i].Free;
      end;
      for i := 0 to 7
      do begin
        combR[ i].Free;
        combL[ i].Free;
      end;
      inherited;
    end;

    function TReverb.getdamp: single;
    begin
      Result := damp / scaledamp;
    end;

    function TReverb.getdry: single;
    begin
      Result := dry / scaledry;
    end;

    function TReverb.getmode: single;
    begin
      if mode >= freezemode
      then Result := 1
      else Result := 0;
    end;

    function TReverb.getroomsize : single;
    begin
      Result := (roomsize - offsetroom) / scaleroom;
    end;

    function TReverb.getwet : single;
    begin
      Result := wet / scalewet;
    end;

    function TReverb.getwidth : single;
    begin
      Result := Width;
    end;

    procedure TReverb.mute;
    var
      i : integer;
    begin
      if getmode >= freezemode
      then
        exit;

      for i := 0 to numcombs - 1
      do begin
        combL[ i].mute;
        combR[ i].mute;
      end;
      for i := 0 to numallpasses - 1
      do begin
        allpassL[ i].mute;
        allpassR[ i].mute;
      end;
    end;

    procedure TReverb.processmix(const inputL : single; const inputR : single; var outputL : single; var outputR : single);
    var
      outL, outR, input : single;
      j                 : integer;
    begin
      outL  := inputL;
      outR  := inputR;
      input := ( inputL + inputR) * gain;
      // Accumulate comb filters in parallel
      for j := 0 to numcombs - 1
      do begin
        outL := outL + combL[ j].process( input);
        outR := outR + combR[ j].process( input);
      end;
      // Feed through allpasses in series
      for j := 0 to numallpasses - 1
      do begin
        outL := allpassL[ j].process( outL);
        outR := allpassR[ j].process( outR);
      end;
      // Calculate output MIXING with anything already there
      outputL := outputL + outL * wet1 + outR * wet2 + inputL * dry;
      outputR := outputR + outR * wet1 + outL * wet2 + inputR * dry;
    end;

    procedure TReverb.processreplace(const inputL : single; const inputR : single; var outputL : single; var outputR : single);
    var
      outL, outR, input : single;
      j                 : integer;
    begin
      outL  := 0;
      outR  := 0;
      input := ( inputL + inputR) * gain;
      // Accumulate comb filters in parallel
      for j := 0 to numcombs - 1
      do begin
        outL := outL + combL[ j].process( input);
        outR := outR + combR[ j].process( input);
      end;
      // Feed through allpasses in series
      for j := 0 to numallpasses - 1
      do begin
        outL := allpassL[ j].process( outL);
        outR := allpassR[ j].process( outR);
      end;
      // Calculate output REPLACING anything already there
      outputL := outL * wet1 + outR * wet2 + inputL * dry;
      outputR := outR * wet1 + outL * wet2 + inputR * dry;
    end;

    procedure TReverb.setdamp( Value : single);
    begin
      damp := Value * scaledamp;
      update;
    end;

    procedure TReverb.setdry( Value : single);
    begin
      dry := Value * scaledry;
      update;
    end;

    procedure TReverb.setmode( Value : single);
    begin
      mode := Value;
      update;
    end;

    procedure TReverb.setroomsize( Value : single);
    begin
      roomsize := ( Value * scaleroom) + offsetroom;
      update;
    end;

    procedure TReverb.setwet( Value : single);
    begin
      wet := Value * scalewet;
      update;
    end;

    procedure TReverb.setwidth( Value : single);
    begin
      Width := Value;
      update;
    end;

    procedure TReverb.update;
    var
      i : integer;
    begin
      // Recalculate internal values after parameter change
      wet1 := wet * ( Width / 2 + 0.5);
      wet2 := wet * (( 1 - Width) / 2);

      if mode >= freezemode
      then begin
        roomsize1 := 1;
        damp1     := 0;
        gain      := muted;
      end
      else begin
        roomsize1 := roomsize;
        damp1     := damp;
        gain      := fixedgain;
      end;

      for i := 0 to numcombs - 1
      do begin
        combL[ i].setfeedback( roomsize1);
        combR[ i].setfeedback( roomsize1);
      end;

      for i := 0 to numcombs - 1
      do begin
        combL[ i].setdamp( damp1);
        combR[ i].setdamp( damp1);
      end;
    end;


end.
